#!/usr/bin/perl -w

=head1 NAME

instantafs.tool - access to various instantafs functions via command line

=head1 SYNOPSIS

instantafs.tool [function] {[function_specific_parameters]}

=head1 DESCRIPTION

instantafs.tool is frontend for various functions of the AFS::Instant
perl library which do not have a dedicated front end command.

=head1 FUNCTIONS

=over 4

=cut

use AFS::Instant::Toolz qw(instantafs_init instantafs_help @AFS_ACL_RIGHTS %AFS_ACL_RIGHTS scanafs dir2vol flipbase64_to_int int_to_flipbase64 make_mount);
use Fb::IO qw(iaexec hash_mkdir writeFile);
use Fb::Data qw(human_readable hash_dir);
use Fb::Network qw(getFQDN);
use Fb::Out qw(out $VERBOSE);
use AFS::Instant::Manage;
use AFS::Instant::CellDB;
use AFS::Instant::Output::Partitions;
use AFS::Instant::Cred;
use POSIX ":sys_wait_h";
use strict;
use warnings;
use Data::Dumper;

my $SHOWVERSION=0;
my $NOHUMANREADABLE=0;
my $DRYRUN=0;
my $MAX_CELLS=0;
my $MAX_VOLUMES=0;

=item volumegroups (Alias: vgs)

This function lists all volume groups of the cell. The VGs names are
printed - sorted by name - one per line. 

Exit codes:

=over 4

=item B<0> if at least one volume group exists.

=item B<1> if no volume group exists.

=back

=cut

sub volumegroups {
	my $celldb=AFS::Instant::CellDB->new();
	my $groups=$celldb->listVolumeGroups();
	my $result=1;
	for my $group (sort keys %{$groups}) {
		print "$group\n";
		$result=0;
	}
	return $result;
}

=item volumegroupmember (Aliase: vgm)

Determines the volumes groups the given volume(s) is/are member of.

Exit codes:

=over 4

=item B<0> if the volumes are member of at least one volume group.

=item B<1> if the volumes are not member of any volume group.

=item B<2> if an error occured (i.e. a given volume doesn't exist).

=back

=cut

sub volumegroupmember {
	my ($volumes,%options)=@_;
	my $celldb=AFS::Instant::CellDB->new();
	my @volrefs;
	for my $volume (@{$volumes}) {
		my $volref=$celldb->getVolume($volume);
		if ( ! defined($volref) ) {
			out('e',"Unable to get volume information of volume '$volume'");
			return 2;
		}
	}
	my $memberships=$celldb->getVolumeMemberships(\@volrefs);
	my $found=0;
	for my $group (sort keys %{$memberships}) {
		print "$group\n";
		$found++;
	}
	return ! ($found != 0);;
}



sub HR {
	my ($value,$mode)=@_;
	if ( ! defined($mode) ) {
		$mode='memory';
	}	
	if ( ! defined($value) ) {
		return "?";
	}
	if ( $NOHUMANREADABLE ) {
		return $value;
	} else {
		if ( $mode eq 'memory' ) {
			return human_readable($value,'memory','space' => '','limit' => 9999);
		} elsif ( $mode eq 'integer' ) {
			return human_readable($value,'integer','space' => '','limit' => 9999);
		} elsif ( $mode eq 'timeinterval' ) {
			return human_readable($value,'timeinterval');
		}
	}
}

=item servers (Alias: s)

Prints the records of all known servers of the cell. The show table contains

=over 4

=item B<name>

The server's display name.

=item B<uuid>

The AFS-UUID of the fileserver.

=item B<ips>

The list of interfaces registered in VLDB for the server plus a bit expressing
the server's availability (a B<!> as IP-prefix in case of the addres being unavailable).

=back

=cut

sub servers {
	my $celldb=AFS::Instant::CellDB->new();
	for my $uuid (@{$celldb->serverInfo('uuidlist')}) {
		my $name=$celldb->serverInfo('uuid2name',$uuid);
		my $addrs=$celldb->serverInfo('uuid2addresses',$uuid);
		my $ioaddrs=$celldb->serverInfo('uuid2ioaddresses',$uuid);
		my %ioaddrs;
		for my $addr (@{$ioaddrs}) {
			$ioaddrs{$addr}=1;
		}
		my @addrs;
		for my $addr (@{$addrs}) {
			if ( ! $ioaddrs{$addr} ) {
				$addr = "!$addr";
			}
			push(@addrs,$addr);
		}
		_align(\$name,20);
		my $s="$name $uuid " . join(' ',@addrs);
		print "$s\n";
	}
	return 0;
}

=item cellstorage (Aliase: cs)

This function shows statistical information about the cell.

=over 4

=item B<cell_size> The sum of the cell's fileservers' partitions' sizes.

=item B<cell_free> The free space on the cell's data partitions (taken from vos partionfo)

=item B<cell_used> The used space on the cell's data partitions

=item B<commited> The sum of the cell's volume instances' quotas.

=item B<commited_percent> The percentage of the cell's volume instance's quota sum in relation to the cell's B<size>.

=item B<commited_rw> The size of the cell's RW instances' quota sum (Net storage utilisation without redundancy on the AFS layer).

=item B<volumes> Number of volumes (this explicitely doesn't mean instances) found on the cell's fileservers.

=item B<volumes_size> The sum of the cell's volume instances' sizes.

=item B<volumes_size_rw> The sum of the cell's RW volume instances' sizes.

=item B<volumegroups> Number of volume groups (original AFS term) found on fileservers (a volume being represented by any number of instances in a data partitions means one volume group).

=item B<instances> Number of volume instances on fileservers. Only rw, ro and backup instances count.

=item B<instances_accessible> Number of instances accessible to AFS clients (given they are mounted somewhere).

=item B<instances_accessible_rw>, B<instances_accessible_ro>, B<instances_accessible_backup>: B<instances_accessible> for a specific volume instance type.

=item B<instances_unaccessible> Number of inaccessible volume instances.

=item B<instances_unaccessible_rw>, B<instances_unaccessible_ro>, B<instances_unaccessible_backup>: B<instances_unaccessible> for a specific volume instance type.

=item B<instances_flaged_offline> Number of volume instances marked offline on their fileserver.

=item B<instances_busy> Number of instances being busy (=having a transaction running on them)

=item B<instances_in_need_of_salvage> Guess what...

=item B<instances_no_vldb> Volume instances that were found on fileservers but without any reference in VLDB.

=item B<files> Number of files in the cell

=item B<files_rw>, B<files_ro>, B<files_backup>: Number of B<files> in a given volume instance type.

=back

=cut

sub cellstorage {
	my $celldb=AFS::Instant::CellDB->new();
	
	# Volume/Volume instance statistics
	my %idata;
	for my $t (0,1,2) {
		for my $f ('files','quota','size','online','offline','count') {
			$idata{$t}->{$f}=0;
		}
	}
	my %volumes;
	my $instances=0;
	my $busy_instances=0;my $salvage_instances=0;my $offline_instances=0;
	my %volumegroups=();my $novldb_instances=0;
	for my $i (@{$celldb->getInstances()}) {
		$idata{$i->{'instancetype'}}->{count}++;
		$idata{$i->{'instancetype'}}->{'files'} += $i->{'files'};
		$idata{$i->{'instancetype'}}->{'quota'} += $i->{'quota'};
		$idata{$i->{'instancetype'}}->{'size'} += $i->{'size'};
		if ( $celldb->instanceInfo('accessible',$i) ) {
			$idata{$i->{'instancetype'}}->{'online'}++;
		} else {
			$idata{$i->{'instancetype'}}->{'offline'}++
		}
		$busy_instances += $i->{'isbusy'};
		$salvage_instances += $i->{'needssalvage'};
		$offline_instances += (1-$i->{'online'});
		$volumes{$i->{'volumename'}}=1;
		$volumegroups{"$i->{'serveruuid'}:$i->{'partition'}:$i->{'volumename'}"}=1;
		$instances++;
		if ( ! defined($i->{rostate}) && ( $i->{instancetype} != 2 ) ) {
			$novldb_instances++;
		}
	}
	my $volumes=scalar(keys %volumes);
	my $commited=0;my $accessible_instances=0;my $unaccessible_instances=0;
	my $volumegroups=scalar(keys %volumegroups);%volumegroups=();
	my $files=0;
	for my $t (0,1,2) {
		$commited += $idata{$t}->{'quota'},
		$accessible_instances += $idata{$t}->{'online'};
		$unaccessible_instances += $idata{$t}->{'offline'};
		$files += $idata{$t}->{files};
	}
	
	# Partition statistics
	my $size=0;
	my $free=0;
	my $partitions=0;
	my $online_partitions=0;
	my $offline_partitions=0;
	my $offline_partitions_os=0;
	for my $l (@{$celldb->expandLocationPattern('*/*',format => 'arrayref')}) {
		$partitions++;
		$size += $l->{'partsize'};
		$free += $l->{'partfree'};
		if ( $celldb->partitionInfo('online',$l) ) {
			$online_partitions++;
		} else {
			$offline_partitions++;
			if ( $celldb->serverInfo('uuidonline',$l->{serveruuid}) ) {
				$offline_partitions_os++;
			}
		}
	}

	# Server statistics	
	my $online_servers=0;
	my $offline_servers=0;
	my $servers=0;
	for my $uuid (@{$celldb->serverInfo('uuidlist')}) {
		if ( $celldb->serverInfo('uuidonline',$uuid) ) {
			$online_servers++;
		} else {
			$offline_servers++;
		}
		$servers++;
	}

	my $compc='-';
	if ( $size > 0 ) {
		$compc=int($commited/$size*100);
	}
	print "cell_size " . HR($size) . "
cell_free " . HR($free) . "
cell_used " . HR($size-$free) . "
commited " . HR($commited) . "
commited_percent $compc
commited_rw " . HR($idata{0}->{'quota'}) . "
volumes $volumes
volumegroups $volumegroups
instances $instances
instances_accessible $accessible_instances
instances_accessible_rw $idata{0}->{'online'}
instances_accessible_ro $idata{1}->{'online'}
instances_accessible_backup $idata{2}->{'online'}
instances_unaccessible $unaccessible_instances
instances_unaccessible_rw $idata{0}->{'offline'}
instances_unaccessible_ro $idata{1}->{'offline'}
instances_unaccessible_backup $idata{2}->{'offline'}
instances_flaged_offline $offline_instances
instances_busy $busy_instances
instances_in_need_of_salvage $salvage_instances
instances_no_vldb $novldb_instances
instances_size_rw " . HR($idata{0}->{'size'}) . "
instances_size_ro " . HR($idata{1}->{'size'}) . "
instances_size_backup " . HR($idata{2}->{'size'}) . "
files " . HR($files,'integer') . "
files_rw " . HR($idata{0}->{'files'},'integer') . "
files_ro " . HR($idata{1}->{'files'},'integer') . "
files_backup " . HR($idata{2}->{'files'},'integer') . "
fileservers $servers
fileservers_online $online_servers
fileservers_offline $offline_servers
locations $partitions
locations_online $online_partitions
locations_offline $offline_partitions
locations_offline_on_online_servers $offline_partitions_os
";
	return 1;
}


sub showversion {
	my $version='INSTANTAFSVERSION';
	print "instantafs.tool $version
Written by Frank Burkhardt
";
}

sub afsfind {
	my @dirs=@_;
	my %options=(
		'max_mountdepth' => $MAX_VOLUMES,
		'max_celldepth' => $MAX_CELLS,
		cb_dir => sub { print "$_[0]\n"; }
	);
	
	for my $dir (@dirs) {
		if ( $dir eq '-' ) {
			while ( my $l=<STDIN> ) {
				chomp($l);
				scanafs($l,%options);
			}
		} else {
			scanafs($dir,%options);
		}
	}
	return 0;
}

sub locationgroupdetails {
	my $celldb=AFS::Instant::CellDB->new() || return 0;
	if ( ! $celldb->ensureCellinfo() ) {
		out('e',"Unable to get cell info");
		return 0;
	}
	my $lgs_=$celldb->{cellinfo}->{xc}->find("//ci:locationgroup");
	my %lgs=();
	
	for my $lg_ (@{$lgs_}) {
		$lgs{$lg_->getAttribute('name')}={
			infotime => [0,undef,0],
			partsize => [0,undef,0],
			partfree => [0,undef,0],
			quotasum => [0,undef,0],
			quotasumrw => [0,undef,0],
			sizesumrw => [0,undef,0],
		};
	}
	for my $lg (sort keys %lgs) {
		my $locations=$celldb->getLocationGroup($lg,format => 'arrayref');
		my %servers_=();my %partitions_=();
		for my $location (@{$locations}) {
			for my $field (qw(infotime partfree partsize quotasum quotasumrw sizesumrw)) {
				$lgs{$lg}->{$field}->[0] += $location->{$field};
				if ( ! defined($lgs{$lg}->{$field}->[1]) || $lgs{$lg}->{$field}->[1] > $location->{$field} ) {
					$lgs{$lg}->{$field}->[1]=$location->{$field};
				}
				if ( $lgs{$lg}->{$field}->[2] < $location->{$field} ) {
					$lgs{$lg}->{$field}->[2]=$location->{$field};
				}
			}
			$servers_{$location->{address}}=1;
			$partitions_{$location->{address} . '/' . $location->{partition}}=1;
		}
		$lgs{$lg}->{partitions}=scalar(keys %partitions_);
		$lgs{$lg}->{servers}=scalar(keys %servers_);
	}
	my $time=time();
	if ( $NOHUMANREADABLE ) {
		print "#Locationgroupname Size Free Used Partfree[min] Partfree[max] Quota Quota[RW] Infotime[old] Infotime[new] Partitions Servers\n";
	} else {
		print "#Locationgroupname          Size   Free   Used  PFmin  PFmax      Q    Qrw USEDrw Told Tnew #P #S\n";
	}
	for my $lg (sort keys %lgs) {
		my $r=$lgs{$lg};
		my $size=HR($r->{partsize}->[0]);
		my $free=HR($r->{partfree}->[0]);
		my $free_min=HR($r->{partfree}->[1]);
		my $free_max=HR($r->{partfree}->[2]);
		my $sum=HR($r->{partsize}->[0]-$r->{partfree}->[0]);
		my $id_min='-';
		if ( $r->{infotime}->[1] ) {
			$id_min=HR($time-$r->{infotime}->[1],'timeinterval');
		}
		my $id_max='-';
		if ( $r->{infotime}->[2] ) {
			$id_max=HR($time-$r->{infotime}->[2],'timeinterval');
		}
		my $quotasumrw=HR($r->{quotasumrw}->[0]);
		my $quotasum=HR($r->{quotasum}->[0]);
		my $sizesumrw=HR($r->{sizesumrw}->[0]);
		my $partitions=$r->{partitions};
		my $servers=$r->{servers};
		if ( ! $NOHUMANREADABLE ) {
			_align(\$lg,25);
			_align(\$size,-6);
			_align(\$free,-6);
			_align(\$free_min,-6);
			_align(\$free_max,-6);
			_align(\$sum,-6);
			_align(\$quotasum,-6);
			_align(\$quotasumrw,-6);
			_align(\$sizesumrw,-6);
			_align(\$id_max,-4);
			_align(\$id_min,-4);
			_align(\$partitions,-2);
			_align(\$servers,-2); 
		}
		print "$lg $size $free $sum $free_min $free_max $quotasum $quotasumrw $sizesumrw $id_min $id_max $partitions $servers\n";
	}
}

sub locationgroups {
	my $celldb=AFS::Instant::CellDB->new();
	if ( ! $celldb->ensureCellinfo() ) {
		out('e',"Unable to get cell info");
		return 0;
	}
	my $lgs_=$celldb->{cellinfo}->{xc}->find("//ci:locationgroup");
	for my $lg_ (@{$lgs_}) {
		my $lg=$lg_->getAttribute('name');
		if ( $VERBOSE ) {
			my $locations=$celldb->expandLocationPattern($lg,format => 'arrayref');
			for my $l (@{$locations}) {
				my $servername=$celldb->serverInfo('uuid2name',$l->{serveruuid});
				$lg .= " " . substr($servername,0,3) . "/" . $l->{partition};
			}
		}
		print $lg . "\n";
	}
	return 0;
}

sub selectlocation {
	my $group=shift;
	my @params=@_;
	my $celldb=AFS::Instant::CellDB->new();
	if ( @params % 3 ) {
		out('e',"Selections rules require 3 arguments per rule");
		return 0;
	}
	my @params_=();
	for(my $n=0;$n<=$#params;$n+=3) {
		my %param=(
			'type' => $params[$n],
			'value' => $params[$n+1],
			'pattern' => $params[$n+2]
		);
		push(@params_,\%param);
	}
	
	my $location=$celldb->selectLocation($group,\@params_);
	if ( $location ) {
		print $celldb->serverInfo('uuid2name',$location->{'serveruuid'}) . "/" . $location->{'partition'} . "\n";
		return 1;
	} else {
		out('w',"No location found");
		return 0;
	}
}


sub volhash {
	my @volumes=@_;
	my $celldb=AFS::Instant::CellDB->new();
	my $base="/afs/" . $celldb->{cellname} . "/admin/volumes";
	if ( ! @volumes ) {
		my $volumes=$celldb->listVolumes();
		if ( $volumes ) {
			@volumes=sort keys %{$volumes};
		} else {
			out('e',"Unable to get list of the cell's volumes");
			return 0;
		}
	}
	out('d',"Using hashing directory '$base'/{rw,ro}");
	for my $volume (@volumes) {
		out('v',"Hashing volume $volume");
		my $hash=hash_dir($volume,2);
		if ( hash_mkdir($volume,$base . '/rw',2,dryrun => $DRYRUN) ) {
			my $mountpoint="$base/rw/$hash->[0]/$hash->[1]/$volume";
			make_mount($mountpoint,$volume,rw => 1,dryrun => $DRYRUN);
		}
		if ( hash_mkdir($volume,$base . '/ro',2,dryrun => $DRYRUN) ) {
			my $mountpoint="$base/ro/$hash->[0]/$hash->[1]/$volume";
			make_mount($mountpoint,"$volume.readonly",dryrun => $DRYRUN);
		}
	}
	return 1;
}

sub gethash {
	my @volumes=@_;
	if ( ! @volumes ) {
		out('e',"No volume names given");
	}
	for my $volume (@volumes) {
		out('v',"Hashing volume $volume");
		my $hash=hash_dir($volume,2);
		print "$volume $hash->[0]/$hash->[1]\n";
	}
	return 1;
}

sub checkxml {
	my @data=@_;
	if ( ! @data ) {
		out('e',"No xml schema given");
		return 0;
	}
	my $mode=shift @data;
	if ( ! @data ) {
		out('i',"No files given");
		return 1;
	}
	my $schema;
	if ( $mode eq 'acallconfig' ) {
		$schema=XML::LibXML::RelaxNG->new(location => '/usr/share/doc/instantafs/instantafs.acallconfig.rng');
	} elsif( $mode eq 'cellstate' ) {
		$schema=XML::LibXML::RelaxNG->new(location => '/usr/share/doc/instantafs/instantafs.cellstate.rng');
	} else {
		out('e',"XML schema id '$mode' is unknown.");
		return 0;
	}
	my $success=1;
	for my $data (@data ) {
		my $doc=XML::LibXML->new->parse_file($data);
		print "$doc\n";
		my $result=eval { $schema->validate($doc) };
		$success &= $result;
		if ( ! $result ) {
			out('w',"File '$data' invalid.");
		}
	}
	return $success;
}

###### ab hier korrekte return-codes

sub dirslots {
	my $dir=shift;
	my $success=1;
	if ( ! -d $dir ) {
		out('e',"dirslots(): '$dir' is not a directory");
		return undef;
	}
	if ( ! opendir(DIR,$dir) ) {
		out('e',"Unable to opendir '$dir': $!");
		return undef;
	}
	my $usedslots=0;my $dentries=0;
	while ( my $dentry=readdir(DIR) ) {
		my $slots;
		if ( length($dentry) <= 16 ) {
			$slots=1;
		} else {
			$slots=1+int((length($dentry)-16)/32)+1;
		}
		$usedslots += $slots;
		$dentries += 1;
	}
	closedir(DIR);
	return {
		'usedslots' => $usedslots,
		'dentries' => $dentries,
		'slots' => 64435,
	}
}

sub sysid {
	if ( scalar(@_) < 3 ) {
		out('e',"Usage: $0 sysid [sysid] [file] [ip] {[ip] {...}}");
	}
	my ($sysid,$file,@ips)=@_;
	if ( $sysid !~ /^[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{2}-[a-f0-9]{2}-[a-f0-9]{12}$/ ) {
		out('e',"Sysid format is invalid.");
		return 0;
	}
	$sysid =~ s/-//g;
	my @content=(0xcc,0xbb,0xaa,0x88,0x01,0x00,0x00,0x00);
	for(my $n=0;$n<16;$n++) {
		push(@content,hex(substr($sysid,$n*2,2)));
	}
	push(@content,scalar(@ips),0,0,0);
	for my $ip (@ips) {
		if ( $ip !~ /^([0-9]{1,3})\.([0-9]{1,3})\.([0-9]{1,3})\.([0-9]{1,3})$/ ) {
			out('e',"Invalid IP address format '$ip'");
			return 0;
		}
		my @ip=($1,$2,$3,$4);
		for my $digit (@ip) {
			if ( $digit < 255 ) {
				push(@content,$digit);
			} else {
				out('e',"Invalid octet '$digit' in ip '$ip'");
				return 1;
			}
		}
	}
	my $content=pack('C' .int(28+4*scalar(@ips)),@content);
	return ! writeFile($file,$content);
}

sub namei2vol {
	my ($celldb,$dir)=@_;
	if ( $dir =~ /^(?:(?:\/vicep[a-z]{1,2}\/AFSIDat\/)?[+=0-9A-Za-z]+\/)?([+=0-9A-Za-z]+)(?:\/.*)?$/ ) {
		my $codedid=$1;
		my $id=flipbase64_to_int($codedid);
		my $data=$celldb->resolveVolumeId($id);
		if ( defined($data) ) {
			my $volume=$data->[0];
			print "$codedid $id $volume\n";
			return 0;
		} else {
			print "$codedid $id (unknown)\n";
			return 1;
		}
	} else {
		out('w',"Unable to parse namei-volume-path '$dir'");
		return 1;
	}
}

sub vol2namei {
	my ($celldb,$vol)=@_;
	my $volid=$vol;
	if ( $volid !~ /^[0-9]+$/ ) {
		my $volref=$celldb->getVolume($vol);
		if ( ! defined($volref) ) {
			out('w',"Unknown volume '$vol'");
			return 0;
		}
		$volid=$volref->{'id'};
	} else {
		my $data=$celldb->resolveVolumeId($volid);
		if ( defined($data) ) {
			my $volref=$celldb->getVolume($vol);
			if ( $volref ) {
				$vol=$volref->{'id'};
			} else {
				out('e',"Unable to resolve volume name to a volume ID");
			}
		} else {
			out('e',"Unable to resolve ID to a volume instance");
		}
	}
	my $namei1=int_to_flipbase64($volid & 255);
	my $namei2=int_to_flipbase64($volid);
	print "$vol $volid $namei1/$namei2\n";
	return 0;
}

sub getvoldir {
	my @volumes=@_;
	my $celldb=AFS::Instant::CellDB->new();
	my $success=1;
	if ( @volumes ) {
		for my $vol (@volumes) {
			$success &= vol2namei($celldb,$vol);
		}
	} else {
		while ( my $vol=<STDIN> ) {
			chomp($vol);
			$success &= vol2namei($celldb,$vol);
		}
	}
	return ! $success;
}

sub getdirvol {
	my @directories=@_;
	my $celldb=AFS::Instant::CellDB->new();
	my $success=1;
	if ( @directories ) {
		for my $dir (@directories) {
			$success &= namei2vol($celldb,$dir);
		}
	} else {
		while ( my $dir=<STDIN> ) {
			chomp($dir);
			$success &= namei2vol($celldb,$dir);
		}
	}
	return ! $success;
}

sub _align {
	my ($s,$a)=@_;
	my $l=length($$s);
	if ( $a > 0 ) {
		if ( $l < $a ) {
			$$s = $$s . (" " x ($a-$l));
		} elsif ( $l > $a )  {
			$$s = substr($$s,0,$a);
		}
	} elsif ( $a < 0 ) {
		$a=abs($a);
		if ( $l < $a ) {
			$$s=(" " x ($a-$l)) . $$s;
		}
	}
}


sub printdir2vol {
	my @directories=@_;
	my $success=1;
	for my $dir (@directories) {
		my $v=dir2vol($dir);
		my $d=dirslots($dir);
		my ($quota,$size,$partsize,$partfree,$volumeid,$cell,$volume,$instance,$dentries,$usedslots)=
			('?','?','?','?','?','?','?','?','?','?');
		$size=HR($v->{'size'});
		if ( defined($v) ) {
			if ( defined($v->{quota}) ) {
				$quota=HR($v->{'quota'});
			} else {
				$quota='-';
			}
			$volume=$v->{'volume'};
			$instance=$v->{'instance'};
			$cell=$v->{'cell'};
			$volumeid=$v->{'id'};
		}
		if ( defined($d) ) {
			$dentries=$d->{'dentries'};
			$usedslots=$d->{'usedslots'};
		}
		$partsize=HR($v->{'partsize'});
		$partfree=HR($v->{'partfree'});
		if ( ! $NOHUMANREADABLE ) {
			_align(\$quota,-6);
			_align(\$size,-6);
			_align(\$partfree,-6);
			_align(\$partsize,-6);
			_align(\$volume,21);
			_align(\$instance,10);
			_align(\$cell,15);
			_align(\$usedslots,-5);
			_align(\$dentries,-5);
		}
		print "$volume $instance $cell $size $quota $partfree $dentries $usedslots $dir\n";
	}
	return ! $success;
}

sub removevolumes {
	my @volumes=@_;
	my $celldb=AFS::Instant::CellDB->new();
	my $manage=AFS::Instant::Manage->new(celldb => $celldb);
	my $success=1;
	for my $volume (@volumes) {
		my $volref=$celldb->getVolume($volume);
		my $vsuccess=1;
		out('v',"Removing volume '$volume'");
		if ( $volref ) {
			for my $iref (@{$celldb->iterateInstances($volref,types => {0 => 1,1 => 1},realonly => 1)}) {
				my $isuccess=$manage->removeInstance($iref,dryrun => $DRYRUN);
				$vsuccess &= $isuccess;
			}
		}
		$success &= $vsuccess;
	}
	return ! $success;
}

sub renamevolumes {
	my @volumes=@_;
	my @renames=();
	if ( ( @volumes >= 2 ) and ( ( @volumes % 2 ) == 0 ) ) {
		if ( ( $volumes[0] !~ /=/ ) and ( $volumes[1] !~ /=/ ) ) {
			for(my $n=0;$n<(@volumes/2);$n++) {
				push(@renames,[$volumes[2*$n],$volumes[2*$n+1]]);
			}
		}
	}
	if ( ! @renames ) {
		for my $r (@renames) {
			if ( $r =~ /^(\S+)=(\S+)$/ ) {
				push(@renames,[$1,$2]);
			} else {
				out('e',"Unable to interpret '$r' as a volume rename rule,aborting");
				return 0;
			}
		}
	}
	my $celldb=AFS::Instant::CellDB->new();
	my $manage=AFS::Instant::Manage->new(celldb => $celldb);
	my $success=1;
	if ( @renames ) {
		for my $r (@renames) {
			$success &= $manage->renameVolume($r->[0],$r->[1],dryrun => $DRYRUN);
		}
	} else {
		out('i',"Enter [oldvolumename]=[newvolumename] pairs - once per line. Ctrl+D to stop");
		while ( my $l=<STDIN> ) {
			chomp($l);
			if ( $l =~ /^(\S+)[\s=,](\S+)$/ ) {
				$success &= $manage->renameVolume($1,$2,dryrun => $DRYRUN);
			}
		}
	}
	return ! $success;
}

sub removeinstances {
	my @instances=@_;
	my $celldb=AFS::Instant::CellDB->new();
	my $manage=AFS::Instant::Manage->new(celldb => $celldb);
	my $success=1;
	for my $instance (@instances) {
		my $irec=$celldb->getInstance($instance);
		if ( $irec ) {
			$success &= $manage->removeInstance($irec,dryrun => $DRYRUN);
		} else {
			out('e',"Unknown volume instance '$instance'");
			$success=0;
		}
	}
	return ! $success;
}

sub addrwroclones {
	my @volumes=@_;
	my $celldb=AFS::Instant::CellDB->new();
	my $manage=AFS::Instant::Manage->new(celldb => $celldb);
	my $success=1;
	for my $volume (@volumes) {
		my $volref=$celldb->getVolume($volume);
		if ( $celldb->volumeInfo('hasrw',$volref) ) {
			if ( ! $celldb->volumeInfo('hasrwro',$volref) ) {
				$success &= $manage->createInstanceRo($volume,$volref->{'rw'},dryrun => $DRYRUN);
			} else {
				out('d',"RWRO already exists for volume '$volume', none added");
			}
		}
	}
	return ! $success;
}

sub removerwroclones {
	my @volumes=@_;
	my $celldb=AFS::Instant::CellDB->new();
	my $manage=AFS::Instant::Manage->new(celldb => $celldb);
	my $success=1;
	for my $volume (@volumes) {
		my $volref=$celldb->getVolume($volume);
		if ( $celldb->volumeInfo('hasrw',$volref) ) {
			my $rwro=$celldb->volumeInfo('rwro',$volref);
			if ( defined($rwro) ) {
				$success &= $manage->removeInstance($rwro,dryrun => $DRYRUN);
			} else {
				out('d',"No RWRO exists for volume '$volume', non removed");
			}
		}
	}
	return ! $success;
}

sub movelocation {
	my ($source,$target)=@_;
	#####
}

sub moveinstances {
	my @instances_=@_;
	my $targetlocation=pop(@instances_);
	my $celldb=AFS::Instant::CellDB->new();
	my $manage=AFS::Instant::Manage->new(celldb => $celldb);
	my $locref=$celldb->expandLocationPattern($targetlocation,format => 'arrayref');
	if ( ! defined($locref) || ! @{$locref} ) {
		out('e',"Unknown location '$targetlocation'");
		out('i',"Did you give the target location as last argument");
		return 2;
	} elsif ( @{$locref} > 1 ) {
		out('e',"Location expression '$targetlocation' resolved to multiple locations");
		return 2;
	}
	$locref=$locref->[0];
	my @instances;
	for my $instance_ (@instances_) {
		my $iref=$celldb->getInstance($instance_);
		if ( $iref ) {
			push(@instances,$iref);
		} else {
			out('e',"Unknown volume instance '$instance_'");
		}
	}
	if ( @instances_ != @instances ) {
		return 1;
	}
	my $errors=0;
	for my $iref (@instances) {
		if ( ! $manage->moveInstance($iref,$locref,auth => 'token',dryrun => $DRYRUN) ) {
			out('e',"Error moving '" . $celldb->instanceInfo('hrname',$iref) . "'");
			$errors++;
		}
	}
	return ! ($errors == 0);
}

sub moveprimary {
	my @volumenames=@_;
	my $targetlocation=pop(@volumenames);
	my $celldb=AFS::Instant::CellDB->new();
	my $manage=AFS::Instant::Manage->new(celldb => $celldb);
	my $locref=$celldb->expandLocationPattern($targetlocation,format => 'arrayref');
	if ( ! defined($locref) || ! @{$locref} ) {
		out('e',"Unknown location '$targetlocation'");
		return 2;
	} elsif ( @{$locref} > 1 ) {
		out('e',"Location expression '$targetlocation' resolved to multiple locations");
		return 2;
	}
	$locref=$locref->[0];
	my @instances;
	my $errors=0;
	for my $volumename (@volumenames) {
		my $instances=$celldb->getInstances(volume => $volumename);
		if ( $instances ) {
			my $rwuuid;
			my $rwpartition; # Just to be safe. It should not really be necessary since AFS doesn't like "cross device links"
			for my $instance (@{$instances}) {
				if ( $instance->{instancetype} == 0 ) {
					$rwuuid=$instance->{serveruuid};
					$rwpartition=$instance->{partition};
				}
			}
			if ( defined($rwuuid) ) {
				# Add the whole "volume group" (AFS term) of the RW instance
				push(@instances,
					grep { $_->{serveruuid} eq $rwuuid && $_->{partition} eq $rwpartition && $_->{instancetype} <= 1 }
					sort { $a->{instancetype} <=> $b->{instancetype} }
					@{$instances}
				);
			} else {
				out('w',"Didn't find a RW instance for volume '$volumename', ignored.");
				$errors++;
			}
		} else {
			out('w',"Unknown volume name '$volumename', ignored.");
			$errors++;
		}
	}
	for my $instance (@instances) {
		if ( ! $manage->moveInstance($instance,$locref,auth => 'token',dryrun => $DRYRUN) ) {
			out('e',"Error moving '" . $celldb->instanceInfo('hrname',$instance) . "'");
			$errors++;
		}
	}
	return ! ($errors == 0);
}

sub listtickets {
	my $cred=AFS::Instant::Cred->new();
	my $result=$cred->listTickets();
	if ( ! defined($result) ) {
		out('e',"Error reading ticket list");
		return 1;
	}
	for my $ticket (@{$result}) {
		print "$ticket\n";
	}
	return 0;
}

sub showtickets {
	my @tickets=@_;
	my $cred=AFS::Instant::Cred->new();
	my $success=1;
	for my $ticket (@tickets) {
		my $result=$cred->getTicket($ticket);
		if ( ! defined($result) ) {
			$success=0;
			next;
		}
		print Dumper($result);
	}
	return ! $success;
}

sub fileclients {
	my $file=shift;
	my $maxparallel=40; # Maximum number of parallel cmdebug processes
	my $timeout=20; # Tiemout in seconds for cmdebug processes
	my $output=iaexec(
		['/usr/bin/fs','examine',$file],
		'readpipe',
		stderr => 'hide',
		format => 'arrayref',
		description => "Reading file ID",
	) || return 1;
	my $fid;
	for my $l (@{$output}) {
		if ( $l =~ /^File .*? \(([0-9\.]+)\) contained in volume [0-9]+$/ ) {
			$fid=$1;
		}
	}
	if ( ! defined($fid) ) {
		out('e',"Unable to find fid of file '$file'");
		return 1;
	}
	$output=iaexec(
		['/usr/bin/fs','where',$file],
		'readpipe',
		stderr => 'hide',
		format => 'string',
		description => "Reading fileserver of given file",
	) || return 1;
	chomp($output);
	my $fileserver;
	if ( $output =~ /is on host (\S+)\s*$/ ) {
		$fileserver=$1;
	}
	if ( ! defined($fileserver) ) {
		out('e',"Unable to find fileserver of file '$file'");
		return 1;
	}
	$output=iaexec(
		['/usr/bin/rxdebug',$fileserver,'-allconnections'],
		'readpipe',
		stderr => 'hide',
		format => 'arrayref',
		description => "Reading connection list of fileserver '$fileserver'",
	) || return 1;
	my %clients;
	for my $l (@{$output}) {
		if ( $l =~ /^Connection from host ([0-9\.]+),/ ) {
			$clients{$1}=1;
		}
	}
	for my $client (keys %clients) {
		my $fqdn=getFQDN($client);
		if ( ! $fqdn ) {
			out('v',"Unable to resolve fqdn for '$client'");
			$fqdn=$client;
		}
		$clients{$client}=$fqdn;
	}
	my %pid2client;my @clients=sort keys %clients;
	my $success=1;my $idles=0;
	my %pid2time;
	while ( scalar(@clients) || scalar(keys %pid2client) ) {
		my $didsomething=0;
		if ( @clients && scalar(keys %pid2client) < $maxparallel ) {
			$didsomething=1;
			my $client=shift @clients;
			my $pid=fork();
			if ( $pid ) {
				$pid2client{$pid}=$client;
				$pid2time{$pid}=time();
				# $client2pid{$client}=$pid;
			} else {
				out('v',"Asking client '$clients{$client}' for callbacks");
				my $cmdebug=iaexec(
					['/usr/bin/cmdebug',$client,'-callbacks'],
					'readpipe',
					stderr => 'hide',
					format => 'arrayref',
					description => "Reading cache entries of afs client '$clients{$client}'",
				) || return 2;
				my $correctfid=0;
				for my $l (@{$cmdebug}) {
					if ( $l =~ /^\*\*.*for 1.([0-9\.]+) / ) {
						$correctfid=int($fid eq $1);
					}
					if ( $correctfid && $l =~ /^\s+([0-9]+) opens/ ) {
						my $opencount=$1;
						print "$clients{$client} $opencount\n";
						exit 0;
					}
				}
				exit 1;
			}
		}
		my $pid=waitpid(-1,WNOHANG);
		if ( $pid > 0 ) {
			$didsomething=1;
			my $client=$pid2client{$pid};
			my $clientname=$clients{$client};
			my $result=$? >> 8;
			if ( ! $result ) {
				out('v',"Received data from '$clientname'");
			} elsif ( $result == 1 ) {
				# No matching fids
			} else {
				out('w',"Error receiving data from '$clientname'");
				$success=0;
			}
			delete $pid2client{$pid};
			delete $pid2time{$pid};
		}
		my $now=time();
		if ( ! $didsomething ) {

			# Kill timedout cmdebug processes
			for my $pid (keys %pid2time) {
				if ( $pid2time{$pid} ) {
					kill(9,$pid);
				} elsif ( ( $now - $pid2time{$pid} ) > $timeout ) {
					kill(15,$pid);
					$pid2time{$pid}=0;
				}
			}

			$idles++;
			sleep 1;
			if ( $idles > 5 ) {
				sleep 2;
			}
			out('d',"'" . scalar(keys %pid2client) . "' cmdebugs running, queue length is '" . scalar(@clients) . "'");
		} else {
			$idles=0;
		}
	}
	return ! $success;
}

sub listpermissions {
	my @p=@_;
	if ( @p ) {
		for my $p (@p) {
			if ( $AFS_ACL_RIGHTS{$p} ) {
				my $r=$AFS_ACL_RIGHTS{$p};
				print "$p	$r->[2]\n";
			}
		}
	} else {
		for my $p (@AFS_ACL_RIGHTS) {
			my $r=$AFS_ACL_RIGHTS{$p};
			print "$p   $r->[0]	$r->[1]\n";
		}
	}
	return 0;
}

sub help {
	print "Usage: iat [command] {[options,parameters]}

Various AFS/InstantAFS-related functions

options:
 -h, --help            This help screen

";
	instantafs_help();

	print "commands:

file namespace operations

  dir2vol|d2v D {...}    Extract information from volume, given directories D ...
                         reside in. Printed information are:
                          volume name, instance type, cell name, volume size, 
                          maximum volume size (quota), free space on partition,
                          number of dentries, number of used afs directory slots,
                          directory name
  fileclients|fc F       Find clients accessing given AFS file F
  listpermissions|lp     List all AFS permissions that can be used in ACLs
  afsfind [dir] {...}    Runs a custom 'find' through afs filespace, returns
                         All directories. Valid options are --max-cells and
                         --max-volumes.
                         Use '-' as directory name to read directories from
                         STDIN.
     --max-volumes=M     Maximum number of volume boundaries to cross with
                         afsfind (default:0)
     --max-cells=M       Maximum number of cell boundaries to cross with
                         afsfind (default:0)

location/volume info:
  
  locationgroups|lgs     List all locationgroups by name
  locationgroupdetails|lgd
                         Show table with locationgroups and details of them
  selectlocation|sl LG R Select a location from locationgroup LG according to
                         selectionrules R.
  volumegroups|vgs       List all known volume groups
  volumegroupmember|vgm V
                         Determine volume group memberships of given volume V

admin

  volhash {[vol] {...}}  Updates /a/a/volumes mountpoint tree. If there are
                         volumes given, their mountpoints will be updated.
                         Use '-' as volume name to read volumes from STDIN
  gethash {[vol] {...}}  Returns hash dirs for given volumes. Use '-' as volume name
                         to read volumes from STDIN.
  vol2namei              Get namei-fileserver-volume directory of given volume
  namei2vol              Get volume id of given namei-fileserver-volume-directory

instance/volume operations

  removevolumes|rv V ... Remove given volumes and all their instances
  removeinstance|ri I ... Removes the given volume instances
  renamevolumes|renv     Rename volumes. Volumes are either given as oldname=newname,
                         as pairs of oldname newname (two parameters per rename)
                         or are given via stdin, one oldname newname per line.
  addrwroclones|aroc     Add RO clones (RWRO,cheap RO instance) for given volumes
  removerwroclones|rroc  Remove RO clones (RWRO,cheap RO instance) of given volumes
  moveinstances|mi I {...} L
                         Move given instances I to location L. Format of I is
                         [Volume]{.[Instancetype]}{\@[Location]}.
  moveprimary|mp V {...} L
                         Move RW and all COW-copies of it to location L. Vs
                         are simple volumesnames.
* movelocation|ml LS LT  Move all instances on location LS to location LT

statistics, cell information

  cellstorage|cs         Shows storage usage information of the cell
  servers|s              Shows information about fileservers and their availability
  sysid SYSID FILE IP {IP {...}}
                         Write a sysid file

credentials

  listtickets|lt         List tickets in kerberos ticket cache
  showticket|st T {...}  Show given tickets

options

 -N, --no-human-readable Show values without decimal conversion.
 -n, --dry-run           Don't actually do something, just show what would've been done.
 
 *: not implemented, yet
";  
}


sub main {
	if ( @ARGV < 1 ) {
		out('e',"No command given. You need --help\n");
		exit 1;
	}
	my $help=0;
	instantafs_init(
		'cmdoptions' => {
			'help|h' => \$help,
			'version|V' => \$SHOWVERSION,
			'no-human-readable|N' => \$NOHUMANREADABLE,
			'dry-run|n' => \$DRYRUN,

			'max-volumes=i' =>\$MAX_VOLUMES,
			'max-cells=i' => \$MAX_CELLS,
		}
	) || exit 1;
	if ( $SHOWVERSION ) {
		showversion();
		exit 0;
	}
	if ( $help or ! @ARGV ) {
		help();
		exit 0;
	}
	my $cmd=shift(@ARGV);
	my %params=(
		'cellstorage|cs' => \&cellstorage,
		'servers|s' => \&servers,
		'afsfind' => \&afsfind,
		'volhash' => \&volhash,
		'gethash' => \&gethash,
		'vol2namei' => \&getvoldir,
		'namei2vol' => \&getdirvol,
		'checkxml|cx' => \&checkxml,
		'locationgroups|lgs' => \&locationgroups,
		'locationgroupdetails|lgd' => \&locationgroupdetails,
		'selectlocation|sl' => \&selectlocation,
		'volumegroups|vgs' => \&volumegroups,
		'volumegroupmember|vgm' => \&volumegroupmember,
		'moveinstances|mi' => \&moveinstances,
		'moveprimary|mp' => \&moveprimary,
		'removevolumes|rv' => \&removevolumes,
		'removeinstances|ri' => \&removeinstances,
		'renamevolumes|renv' => \&renamevolumes,
		'addrwroclones|aroc' => \&addrwroclones,
		'removerwroclones|rroc' => \&removerwroclones,
		'sysid' => \&sysid,
		'dir2vol|d2v' => \&printdir2vol,
		'movelocation|ml' => \&movelocation,
		'listtickets|lt' => \&listtickets,
		'showtickets|st' => \&showtickets,
		'fileclients|fc' => \&fileclients,
		'listpermissions|lp' => \&listpermissions,
	);
	my $function;
	foreach my $param (keys %params) {
		if ( $cmd =~ /^($param)$/ ) {
			$function=$params{$param};
		}
	}
	if ( defined($function) ) {
		exit &{$function}(@ARGV);
	}
	out('e',"Unknown or command or no command given. You need --help");
	exit 1;
}


main();
